#ifndef __SCENE_HPP__
#define __SCENE_HPP__

#include <algorithm>
#ifdef WIN32
#include <functional>
#else
#include <tr1/functional>
#endif

/// typy obiektow gry
enum objectTypes
{
    OBJECT_TYPE_UNSET = 0,  //obiekt niezainicjalizowany
    OBJECT_TYPE_PLANE,      //plaszczyzna tworzaca plansze
    OBJECT_TYPE_WALL_N,     //sciana polnocna wyznaczajaca granice planszy
    OBJECT_TYPE_WALL_EW,    //sciana wsch oraz zach wyznaczajaca granice planszy
    OBJECT_TYPE_BALL,       //kula
    OBJECT_TYPE_PADDLE,     //rakietka za pomoca ktorej gracz odbija kule
    OBJECT_TYPE_BRICK,      //cegla ktora nalezy rozbic za pomoca kuli

    OBJECT_TYPE_COUNT       //liczba wszystkich zdefiniowanych obiektow
};


/**
 * @struct baseSceneObject
 * @brief struktura zawierajaca podstawowe informacje o obiekcie
 */
struct baseSceneObject
{
    /**
     * @brief domyslny konstruktor
     */
    baseSceneObject( void ) :
              position( vector3f( .0f, .0f, .0f ) )
            , objectType( OBJECT_TYPE_UNSET )
            , selfId( -1 )
    { }

    /**
     * @brief konstruktor ustawiajacy podane wartosci
     * @param pos pozycja obiektu
     * @param objType typ obiektu
     * @param matId identyfikator materialu
     */
    explicit baseSceneObject(
          int32_t objId
        , const vector3f& pos
        , const vector3f& _dir
        , float vel
        , objectTypes objType ) :
              position( pos )
            , objectType( objType )
            , dir( _dir )
            , velocity( vel )
            , selfId( objId )
    { }

    /**
     * @brief zwraca nadany wcześniej identyfikator
     */
    int32_t getId( void ) const
    {
        return selfId;
    }

    /// pozycja obiektu
    vector3f    position;

    /// typ obiektu
    objectTypes objectType;

    /// kierunek poruszania sie aktora
    vector3f    dir;

    /// jego predkosc
    float       velocity;

private:
    /// prywatny identyfikator obiektu
    int32_t     selfId;
};


/**
 * @class sceneManager
 */
template< typename T >
class sceneManager
{
    public: // public typedefs
        typedef std::map< int32_t, baseSceneObject > sceneObjectsContainer;

    public:
        /**
         * @brief Domyslny konstruktor
         * Inicjalizuje materialy i siatki aktorow
         */
        sceneManager( void ) : ids( 0 )
        {
            initializeMaterials();
            initializeMeshes();
        }

        /**
         * @brief Dodaje do sceny obiekt o podanym typie
         * @param objType
         * @param pos
         * @return identyfikator nadany obiektowi
         */
        int32_t addSceneObject(
              const objectTypes& objType
            , const vector3f& pos
            , const vector3f& dir = vector3f()
            , float vel = 0.0f
            , const vector3f& orient = vector3f() )
        {
            int32_t newId = ids++;

            //warunek wstępny
            assert( sceneObjects.find( newId ) == sceneObjects.end() );

            sceneObjects.insert( sceneObjectsContainer::value_type( newId
                , baseSceneObject
                  (
                      newId
                    , pos
                    , dir
                    , vel
                    , objType
                  ) ) );

            return newId;
        }

        /**
         * @brief Usuwa ze sceny obiekt o podanym id
         * @param objId
         */
        void removeSceneObject( int32_t objId )
        {
            sceneObjectsContainer::iterator it = sceneObjects.find( objId );

            if( it != sceneObjects.end() )
            {
                sceneObjects.erase( it );
            }
            else
            {
                std::ostringstream e;
                e << "Scene object of id=";
                e << objId;
                e << " does not exists in scene container.";
                throw std::logic_error( e.str() );
            }
        }

        /**
         * @brief Wyswiela wszystkich aktualnie przechowywanych w scenie aktorow
         */
        void renderAllObjects( void )
        {
            std::for_each( sceneObjects.begin(), sceneObjects.end(),
                std::tr1::bind(
                      &sceneManager::renderObject
                    , this
                    , std::tr1::placeholders::_1 ) );
        }

        /**
         * @brief Wywoluje podany w parametrze szablonu funktor
         * dla kazdego aktora w scenie w celu jej aktualizacji.
         */
        template< typename updaterType >
        void updateScene( void )
        {
            updaterType updater;
            std::for_each( sceneObjects.begin(), sceneObjects.end(),
                std::tr1::bind( &updaterType::operator(), &updater, std::tr1::placeholders::_1 ) );
        }

        /**
         * @brief Wywoluje podany funktor dla kazdego elementu w scenie
         * w celu sprawdzenia kolizji miedzy obiektami.
         */
        template< typename collSolver >
        void solveCollisions( void )
        {
            collSolver solver;
            const sceneObjectsContainer& objects = sceneObjects;

            std::for_each( sceneObjects.begin(), sceneObjects.end(),
                std::tr1::bind(
                      &collSolver::operator()
                    , &solver, std::tr1::placeholders::_1
                    , this
                    , std::tr1::cref( objects ) ) );
        }

    private:
        /**
         * @brief Metoda pomocnicza wyswietla pojedynczego aktora
         * @param value krotka zawierajacego aktora ktory ma byc wyswietlony
         */
        void renderObject( const sceneObjectsContainer::value_type& value )
        {
            const baseSceneObject& object = value.second;

            glPushMatrix();

            glTranslatef( object.position.x, object.position.y, object.position.z );

            glVertexPointer( 3, GL_FLOAT, sizeof( vertexes::value_type ), &objectsMesh[ object.objectType ][ 0 ] );
            glNormalPointer( GL_FLOAT, sizeof( vertexes::value_type ), &objectsMesh[ object.objectType ][ 0 ].normal );

            setMaterial(
                  &objectsMaterial[ object.objectType ].ambient_col.x
                , &objectsMaterial[ object.objectType ].diffuse_col.x
                , &objectsMaterial[ object.objectType ].specular_col.x
                , objectsMaterial[ object.objectType ].shiness );

            glDrawArrays( GL_QUADS, 0, objectsMesh[ object.objectType ].size() );

            glPopMatrix();
        }

        /**
         * @brief Metoda pomocnicza, wypelnia tablice materialow
         */
        void initializeMaterials( void )
        {
            objectsMaterial[ OBJECT_TYPE_PLANE ] = T::planeMaterial;
            objectsMaterial[ OBJECT_TYPE_WALL_N ] = T::wallNMaterial;
            objectsMaterial[ OBJECT_TYPE_WALL_EW ] = T::wallEWMaterial;
            objectsMaterial[ OBJECT_TYPE_BALL ] = T::ballMaterial;
            objectsMaterial[ OBJECT_TYPE_PADDLE ] = T::paddleMaterial;
            objectsMaterial[ OBJECT_TYPE_BRICK ] = T::brickMaterial;
        }

        /**
         * @brief Metoda pomocnicza wypelnia tablice siatek
         */
        void initializeMeshes( void )
        {
            objectsMesh[ OBJECT_TYPE_PLANE ] = createPlane( T::planeWidth, T::planeHeight );
            objectsMesh[ OBJECT_TYPE_WALL_N ] = createRectangular( T::wallNWidth, T::wallNHeight, T::wallNThickness );
            objectsMesh[ OBJECT_TYPE_WALL_EW ] = createRectangular( T::wallEWWidth, T::wallEWHeight, T::wallEWThickness );
            objectsMesh[ OBJECT_TYPE_BALL ] = createSphere( T::ballRadius, T::ballSlices, T::ballSegments );
            objectsMesh[ OBJECT_TYPE_PADDLE ] = createRectangular( T::paddleWidth, T::paddleHeight, T::paddleThickness );
            objectsMesh[ OBJECT_TYPE_BRICK ] = createRectangular( T::brickWidth, T::brickHeight, T::brickThickness );
        }

    private:
        /// identyfikatory obiektow tworzone z wartości przechowywanej
        /// w tej zmiennej
        int32_t ids;

        /// kontener do przechowywania obiektów sceny - aktorow
        sceneObjectsContainer sceneObjects;

        /// tutaj bedziemy pamietac material dla kazdego typu obiektow
        material objectsMaterial[ OBJECT_TYPE_COUNT ];

        ///
        vertexes objectsMesh[ OBJECT_TYPE_COUNT ];
};


#endif // __SCENE_HPP__
